import re

__all__ = ["Record"]


class Record(dict):
    def __repr__(self):
        return self._repr_with_visited(set())

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            ) from None

    def _repr_with_visited(self, visited):
        # Handle cycles by checking if this object is already being processed
        obj_id = id(self)
        if obj_id in visited:
            return "{...}"

        visited.add(obj_id)
        try:
            if not self:
                return "{}"

            items = []
            for key, value in self.items():
                key_str = self._format_key(key)

                # Handle value representation with cycle detection
                if isinstance(value, Record):
                    value_str = value._repr_with_visited(visited.copy())
                elif hasattr(value, "__dict__") and hasattr(value, "__class__"):
                    # For other objects that might contain cycles, use a simple repr
                    value_str = repr(value)
                else:
                    value_str = repr(value)

                items.append(f"{key_str}: {value_str}")

            return "{" + ", ".join(items) + "}"
        finally:
            visited.remove(obj_id)

    def _format_key(self, key):
        if isinstance(key, str) and self._is_identifier(key):
            # If key is an identifier, drop the quotes
            return key

        elif isinstance(key, (int, float, bool, type(None))):
            # If key is a literal like 0, 1, True, False, None, use repr
            return repr(key)

        else:
            # Otherwise, use f"({repr(key)})"
            return f"({repr(key)})"

    def _is_identifier(self, s):
        return (
            isinstance(s, str)
            and re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", s)
            and s not in koatl_keywords
        )


koatl_keywords = {
    "if",
    "else",
    "while",
    "for",
    "def",
    "class",
    "return",
    "import",
    "from",
    "as",
    "with",
    "try",
    "except",
    "finally",
    "raise",
    "assert",
    "async",
    "await",
}



from . import _rs


def transpile(*args, **kwargs):
    try:
        return _rs.transpile(*args, **kwargs)
    except SyntaxError as e:
        raise SyntaxError(e.args[0].decode("utf8")) from None


def transpile_raw(*args, **kwargs):
    try:
        return _rs.transpile_raw(*args, **kwargs)
    except SyntaxError as e:
        raise SyntaxError(e.args[0].decode("utf8")) from None



"""
At the time of writing, I haven't implemented async def
in Koatl, so this is a workaround.

This function converts a generic awaitable (perhaps a
vanilla generator) into a coroutine that can be used
by asyncio.run, which expects strictly a coroutine.

(In practice, async def shouldn't be needed anyways,
since we have the Async monad.)
"""


async def to_coro(awaitable):
    return await awaitable



import pytest
import koatl.cli
from pathlib import Path


def get_test_data():
    data_dirs = [
        Path(__file__).parent / "parse",
    ]

    test_cases = []
    for data_dir in data_dirs:
        for file_path in data_dir.glob("*.tl"):
            test_cases.append(pytest.param(file_path, id=str(file_path)))

    return test_cases


@pytest.mark.parametrize("test_file", get_test_data())
def test_parse(test_file):
    koatl.cli.transpile_from_path(test_file, mode="script")



def transpile_from_source(source, mode="script", script_path="<string>"):
    from koatl import transpile
    import ast

    transpiled_code = transpile(source, mode=mode, filename=str(script_path))

    return ast.unparse(transpiled_code)


def run_from_source(source, mode="script", script_path="<string>"):
    from koatl import transpile

    transpiled_code = transpile(source, mode=mode, filename=str(script_path))
    code_obj = compile(transpiled_code, script_path, "exec")

    script_globals = {"__name__": "__main__"}

    exec(code_obj, script_globals)


def transpile_from_path(script_path, mode="script"):
    with open(script_path, "r") as f:
        original_script_code = f.read()
    transpile_from_source(original_script_code, mode=mode, script_path=script_path)


def run_from_path(script_path, mode="script"):
    with open(script_path, "r") as f:
        original_script_code = f.read()
    run_from_source(original_script_code, mode=mode, script_path=script_path)



from IPython.core.magic import register_line_magic

from koatl import transpile
from .magic import (
    koatl_cell_magic,
    KoatlSystemAssign,
    KoatlMagicAssign,
    KoatlEscapedCommand,
)

import ast


def source_code_transformer(lines):
    source = "".join(lines)
    py_ast = transpile(source, mode="interactive")
    transformed_source = ast.unparse(py_ast)

    lines = [line + "\n" for line in transformed_source.splitlines()]

    return lines


def import_prelude(namespace):
    exec("from koatl.runtime import *", namespace)
    exec("from koatl.prelude import *", namespace)


def load_ipython_extension(ipython):
    import_prelude(ipython.user_ns)
    print("koatl enabled")

    ttm = ipython.input_transformer_manager

    def activate_compiler():
        if source_code_transformer not in ipython.input_transformers_post:
            ipython.input_transformers_post.append(source_code_transformer)

        if koatl_cell_magic not in ttm.line_transforms:
            ttm.line_transforms.insert(0, koatl_cell_magic)

        for cls in [KoatlMagicAssign, KoatlEscapedCommand, KoatlSystemAssign]:
            if cls not in ttm.token_transformers:
                ttm.token_transformers.insert(0, cls)

        ipython.koatl_active = True

    def deactivate_compiler():
        if source_code_transformer in ipython.input_transformers_post:
            ipython.input_transformers_post.remove(source_code_transformer)

        if koatl_cell_magic in ttm.line_transforms:
            ttm.line_transforms.remove(koatl_cell_magic)

        for cls in [KoatlMagicAssign, KoatlEscapedCommand, KoatlSystemAssign]:
            if cls in ttm.token_transformers:
                ttm.token_transformers.remove(cls)

        ipython.koatl_active = False

    @register_line_magic
    def koatl(line):
        """
        Magic command to toggle the custom compiler.
        Usage:
          %koatl on
          %koatl off
          %koatl status
        """
        command = line.strip().lower()
        if command == "on":
            activate_compiler()
        elif command == "off":
            deactivate_compiler()
        elif command == "status":
            status = "active" if ipython.custom_compiler_active else "inactive"
            print(f"koatl is {status}.")
        else:
            print("Usage: %koatl <on|off|status>")

    activate_compiler()


def unload_ipython_extension(ipython):
    if KoatlMagicAssign in ipython.input_transformer_manager.token_transformers:
        ipython.token_transformers.remove(0, KoatlMagicAssign)
    if KoatlSystemAssign in ipython.input_transformer_manager.token_transformers:
        ipython.token_transformers.remove(1, KoatlSystemAssign)
    if source_code_transformer in ipython.input_transformers_post:
        ipython.input_transformers_post.remove(source_code_transformer)

    print("koatl disabled.")
